{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DeriveGeneric #-}

module HAL.Doc.Document where

import Control.Lens qualified as L
import Data.Aeson ( (.:), (.:?), withObject, FromJSON(parseJSON), Object )
import Data.Default (Default(..))
import Data.Map.Strict qualified as Map
import Data.String (IsString(fromString))
import HAL.Utils (allLangs, langAbstractS)
import Data.LanguageCodes (ISO639_1(..))
import Protolude
import Servant.API (ToHttpApiData(..))
import qualified Data.Text as T
import Data.Aeson.Types (Parser)
import Prelude qualified as P

data Document = Document
  { _document_docid                :: Text
  , _document_title                :: [Text]
  , _document_abstract             :: [Text]
  , _document_abstract_lang_map    :: Map ISO639_1 [Text]
  , _document_date                 :: Maybe Text
  , _document_source               :: Maybe Text
  , _document_authors_names        :: [Text]
  , _document_authors_affiliations :: [Text]
  , _document_institutes_tree      :: Map Text [Text]
  , _document_struct_id            :: [Int]
  , _document_original             :: Object
  } deriving (Show, Generic)
L.makeLenses ''Document

instance Default Document where
  def = Document "default Id" def def def def def def def def def mempty

instance FromJSON Document where
  parseJSON = withObject "Document" $ \o -> do
    _document_docid                <- o .: "docid"
    _document_title                <- o .: "title_s" <|> return []
    _document_abstract             <- o .: "en_abstract_s" <|> return []
    _document_date                 <- o .:? "submittedDate_s"
    _document_source               <- o .:? "source_s"
    _document_authors_names        <- o .: "authFullName_s" <|> return []

    idsNames <- o .:? "structIdName_fs" :: Parser (Maybe [Text]) --unsparsed (contains a _FacetSep_)

    let structIdname = getStructIdsNames idsNames

    let _document_struct_id = map fst structIdname
    let _document_authors_affiliations = map snd structIdname


{- 
structIsChildOf_fs is a field in the HAL API that is a list that goes by pair
here is an exemple :
https://api.archives-ouvertes.fr/search/?q=(en_title_t:(glass)%20OR%20en_abstract_t:(glass))%20AND%20(structId_i:6279)%20AND%20(language_s:en)&wt=json&fl=docid,label_s,uri_s,structId_i,title_s,structIsChildOf_fs&rows=80) 

the first line of a pair tells you which type of a structure the child is (e.g. laboratory, institution, etc...)
the second line tells you what is the structId and name of the child structure aswell as the strucId of the parent and the name of the parent structure

for now we are not using the type of structure that's why we have a removeUnusedData that removes the first line of each pair
-}
    structChild <- o .:? "structIsChildOf_fs" :: Parser (Maybe [Text])

    let _document_institutes_tree = Map.fromListWith (++) $ getStructParentChild $ removeUnusedData $ fromMaybe [] structChild

    abstracts <-
      mapM (\lang -> do
               ma <- o .:? fromString (T.unpack $ langAbstractS lang)
               pure $ (\a -> (lang, a)) <$> ma) allLangs
    let _document_abstract_lang_map = Map.fromList $ catMaybes abstracts

    let _document_original = o

    pure $ Document { .. }

-- | Remove one line out two because the field `structIsChildOf_fs` contains
--   informations that is not used for now, the type of the structure (e.g. laboratory, institute, etc...)
-- 
--   Exemple of request : 
--   https://api.archives-ouvertes.fr/search/?q=(en_title_t:(glass)%20OR%20en_abstract_t:(glass))%20AND%20(structId_i:6279)%20AND%20(language_s:en)&wt=json&fl=docid,label_s,uri_s,structId_i,title_s,structIsChildOf_fs&rows=80) 
removeUnusedData :: [a] -> [a]
removeUnusedData (_ : struct : structs) = struct : removeUnusedData structs
removeUnusedData [_] = []
removeUnusedData [] = []

-- | Parses the field structIdName_fs that looks like : 
-- > StructId_FacetSep_StructName
-- 
-- returns [(StructId, StructName)]
getStructIdsNames :: Maybe [T.Text] -> [(Int, T.Text)]
getStructIdsNames (Just idsNames) = map (\tab -> (P.read (T.unpack (P.head tab)) :: Int, P.last tab)) $ splitInstitutes idsNames
  where
    splitInstitutes :: [T.Text] -> [[T.Text]]
    splitInstitutes = P.map (T.splitOn (T.pack "_FacetSep_"))
getStructIdsNames Nothing = []

getStructParentChild :: [T.Text] -> [(T.Text, [T.Text])]
getStructParentChild struct = map (\tab -> (formatParent $ P.last tab, [formatChild $ P.head tab] )) $ splitStruct struct
  where
    splitStruct :: [T.Text] -> [[T.Text]]
    splitStruct = P.map (T.splitOn (T.pack "_JoinSep_"))

    formatChild c = name <> " | " <> id
      where
        name = P.last splited
        id = P.head splited
        splited = T.splitOn "_" c

    formatParent p = name <> " | " <> id
      where
        name = P.last splited
        id = P.head splited
        splited = T.splitOn "_FacetSep_" p

instance ToHttpApiData Document where
  toUrlPiece _ = "docid,title_s,en_abstract_s,fr_abstract_s,submittedDate_s,source_s,authFullName_s,structId_i,structIdName_fs"
